import typing as t

import a2.dataset.load_dataset
import datasets
import numpy as np
import transformers
import xarray


class DatasetHuggingFace:
    """
    Used to create dataset in Hugging Face format.

    Attributes
    ----------
    tokenizer : Hugging Face tokenizer

    Notes
    -----
    Use Hugging Face model folders to initialize tokenizer and build a
    Hugging Face dataset from an xarray dataset.
    Can be reused to bild the test dataset for validation.
    """

    def __init__(self, model_folder: str, use_fast: bool = False):
        self.model_folder = model_folder
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(
            model_folder,
            use_fast=False,
        )

    def _tok_func(self, x: dict):
        return self.tokenizer(x["inputs"], padding=True)

    def build(
        self,
        ds: xarray.Dataset,
        indices_train: np.ndarray | None,
        indices_validate: np.ndarray | None,
        train: bool = True,
        key_inputs: str = "text",
        key_label: str = "raining",
        reset_index: bool = True,
        prediction_dataset: bool = False,
    ) -> t.Union[datasets.Dataset, datasets.DatasetDict]:
        """
        Create Hugging Face dataset (datasets.DatasetDict) from xarray dataset

        Parameters:
        ----------
        ds: xarray dataset
        indices_train: Variable 'index' values that are used for training
        indices_validate: Variable 'index' values that are used for validation
        train: If the dataset is used for training
        key_inputs: Key of variable used as input to model
        key_label: Key of variable used as label for training
        reset_index: Reset index coordinate
        prediction_dataset: Whether to build a dataset used for prediction (no labels required, only added if present).

        Returns
        -------
        datasets.DatasetDict
        """
        if train and (sum([indices_train is None, indices_validate is None]) == 1):
            raise ValueError(f"{indices_train=} and {indices_validate} can either be both None or have to be set!")
        if prediction_dataset:
            required_keys = [key_inputs]
            if key_label in ds:
                required_keys = required_keys + [key_label]
        else:
            required_keys = [key_inputs, key_label]
        a2.dataset.utils_dataset.assert_keys_in_dataset(ds, required_keys)
        if reset_index:
            ds = a2.dataset.load_dataset.reset_index_coordinate(ds)
        if not train and indices_validate is not None:
            ds = ds.sel(index=indices_validate)
        df = ds[required_keys].to_pandas()
        columns: t.Mapping = {key_inputs: "inputs", key_label: "label"}
        df = df.rename(columns=columns, errors="ignore")  # type: ignore
        datasets_ds = datasets.Dataset.from_pandas(df)
        if key_label in required_keys:
            updated_features = datasets_ds.features.copy()
            updated_features["label"] = datasets.ClassLabel(names=[f"not {key_label}", f"{key_label}"])
            datasets_ds = datasets_ds.cast(updated_features)
        tok_ds = datasets_ds.map(self._tok_func, batched=True)
        if not train or (indices_train is None and indices_validate is None):
            return tok_ds
        else:
            return datasets.DatasetDict(
                {
                    "train": tok_ds.select(indices_train),
                    "test": tok_ds.select(indices_validate),
                }
            )



import logging
import math
import os
import pathlib
import time
import typing as t

import a2.utils.file_handling
import requests


def _get_token() -> str:
    """
    Returns twitter api token, which is saved as a global
    variable to authenticate user.
    """
    token = os.getenv("TOKEN")
    if token is None:
        raise ValueError("twitter api token: TOKEN not found as environment variable!")
    return token


def _create_headers(bearer_token: str) -> dict:
    """
    Creates header dict of url for the request based on bearer_token.
    """
    headers = {"Authorization": f"Bearer {bearer_token}"}
    return headers


def create_headers_with_token():
    """
    creates twitter api token and create header
    """
    bearer_token = _get_token()
    return _create_headers(bearer_token)


def _create_url_and_query_parameters(
    keyword: str,
    start_date: t.Optional[str] = None,
    end_date: t.Optional[str] = None,
    max_results: int = 10,
    search_url: str = "https://api.twitter.com/2/tweets/search/all",
) -> tuple[str, dict]:
    """
    Creates url for the request type search all.

    Parameters:
    ----------
    keyword: search keyword
    start_date: start date of tweets
    end_date: end date of tweets
    max_results: maximum number searched for

    Returns
    -------
    tuple(search url, query parameters)
    """

    query_params: dict = {
        "query": keyword,
        "start_time": start_date,
        "end_time": end_date,
        "max_results": max_results,
        "expansions": "author_id,in_reply_to_user_id,geo.place_id",
        "tweet.fields": "id,text,author_id,in_reply_to_user_id,geo,"
        "conversation_id,created_at,lang,public_metrics,"
        "referenced_tweets,reply_settings,source",
        "user.fields": "id,name,username,location,created_at," "description,public_metrics,verified",
        "place.fields": "full_name,id,country,country_code,geo,name," "place_type,contained_within",
        "next_token": None,
    }
    return search_url, query_params


def _connect_to_endpoint(
    url: str,
    headers: t.Mapping,
    params: dict,
    next_token: t.Optional[str] = None,
):
    """
    Connect to endpoint to request search query.

    Requests user query specified in url and header plus
    authentification from header.
    As single requests return a maximum of 500 tweets,
    need to save next token to keep quering for.
    Parameters:
    ----------
    url: url to query
    headers: headers of query
    params: query params (next token will be updated)
    next_token: will update former `next_token` in `params`

    Returns
    -------
    dict of json response
    """
    params["next_token"] = next_token  # params object received from create_url function
    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()
    return response.json()


def _download_tweets(
    headers: dict,
    path_json: t.Union[str, pathlib.Path],
    keyword: str,
    start_dates: t.Sequence[str],
    end_dates: t.Sequence[str],
    max_results: int = 10,
    sleep_time: float = 1.0,
) -> None:
    """
    Download tweets and save specified fields to csv and
    dump all output to json file

    Request tweets from twitter api and save full json response and
    save specified response fields.
    Need to have global variable TOKEN with twitter api token.
    Single requests only return 500 tweets such that multiple queries are
    required per search.
    Queries via twitters 'search all' endpoint
    (see https://developer.twitter.com/en/docs/twitter-api/tweets/search/api-reference/get-tweets-search-all for more details) # noqa
    Parameters:
    ----------
    headers: header dict of url for the request based on bearer_token
    path_csv: path of .csv file
    path_json: path to .json file
    fields: fields to convert from json to csv file to reduce file size
    keyword: keyword specifying query
    start_dates: list of start dates for query in format YYYY-MM-DDTHH:mm:ssZ
    end_dates: list of end dates for query in format YYYY-MM-DDTHH:mm:ssZ
    max_results: number of maximum results
    sleep_time: time in sec waited after individual request, note a maximum of 500 tweets can be obtained per single request

    Returns
    -------

    """
    max_count = math.ceil(max_results / len(start_dates))  # Max tweets per time period
    total_tweets = 0
    for i_dates, (start_date, end_data) in enumerate(zip(start_dates, end_dates)):
        count = 0  # Counting tweets per time period
        stay_this_time = True
        next_token = None
        max_count_request = min([max_count, 500])
        while stay_this_time:
            # Check if max_count reached
            if count >= max_count:
                break
            if count + max_count_request > max_count:
                max_count_request = max([int(max_count - count), 10])
            url, params = _create_url_and_query_parameters(keyword, start_date, end_data, max_count_request)
            json_response = _connect_to_endpoint(url, headers, params, next_token)
            result_count = json_response["meta"]["result_count"]

            _save_response_to_json(json_response, path_json)
            if "next_token" in json_response["meta"]:
                # Save the token to use for next call
                next_token = json_response["meta"]["next_token"]
                if result_count is not None and result_count > 0 and next_token is not None:
                    count, total_tweets = _increment_counters(count, result_count, total_tweets)
                    logging.info(f"Total # of Tweets added: {total_tweets}")
                    time.sleep(sleep_time)
                    # If no next token exists
            else:
                if result_count is not None and result_count > 0:
                    count, total_tweets = _increment_counters(count, result_count, total_tweets)
                    time.sleep(sleep_time)

                # Since this is the final request, turn flag to false
                # to move to the next time period.
                stay_this_time = False
                next_token = None
            time.sleep(sleep_time)
    logging.info(f"Total number of results: {total_tweets}")


def _increment_counters(count: int, result_count: int, total_tweets: int) -> t.Tuple[int, int]:
    count += result_count
    total_tweets += result_count
    return count, total_tweets


def _save_response_to_json(json_response: dict, path_json: t.Union[str, pathlib.Path]):
    a2.utils.file_handling.json_dump(path_json, list(), check_is_file=True)
    data = a2.utils.file_handling.json_load(path_json)
    data.append(json_response)
    a2.utils.file_handling.json_dump(path_json, data)



import collections
import typing as t
from collections.abc import Sequence

import a2.utils.constants
import a2.utils.utils
import matplotlib.colors
import matplotlib.figure
import numpy as np
from matplotlib import pyplot as plt


def set_x_log(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    log: str = "false",
    axis_symlog_linear_threshold: float | None = None,
) -> a2.utils.constants.TYPE_MATPLOTLIB_AXES:
    """
    Sets scale of x-axis

    Parameters:
    ----------
    axes:
        Matplotlib axes
    log:
        Type of bins. Log types included "false", "symlog", "log"
    axis_symlog_linear_threshold:
        Threshold below which bins are linear to include zero values (when `log`="symlog")

    Returns
    -------
    axes
    """
    if log == "symlog":
        if axis_symlog_linear_threshold is None:
            raise ValueError(f"If log=='symlog', setting " f"{axis_symlog_linear_threshold=} required!")
        axes.set_xscale("symlog", linthresh=axis_symlog_linear_threshold)
    elif log == "log":
        axes.set_xscale("log")
    return axes


def set_y_log(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    log: str = "false",
    axis_symlog_linear_threshold: float | None = None,
) -> a2.utils.constants.TYPE_MATPLOTLIB_AXES:
    """
    Sets scale of y-axis

    Parameters:
    ----------
    axes:
        Matplotlib axes
    log:
        Type of bins. Log types included "false", "symlog", "log"
    axis_symlog_linear_threshold:
        Threshold below which bins are linear to include zero values (when `log`="symlog")

    Returns
    -------
    axes
    """
    if log == "symlog":
        if axis_symlog_linear_threshold is None:
            raise ValueError("If log=='symlog', " f"setting: {axis_symlog_linear_threshold=} required!")
        axes.set_yscale("symlog", linthresh=axis_symlog_linear_threshold)
    elif log == "log":
        axes.set_yscale("log")
    return axes


def _create_figure(figure_size):
    return plt.figure(figsize=figure_size)


def set_font(font_size=10):
    SMALL_SIZE = font_size
    MEDIUM_SIZE = font_size
    BIGGER_SIZE = font_size

    plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
    plt.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
    plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
    plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title


def set_title(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    title: str | None,
    font_size: float = a2.utils.constants.DEFAULT_FONTSIZE_SMALL,
):
    if title is not None:
        if isinstance(title, str) and len(title) > 0:
            axes.set_title(title, fontdict={"fontsize": font_size})


def set_axes_label(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    label: str | None,
    axis: str = "x",
    font_size: float = a2.utils.constants.DEFAULT_FONTSIZE_SMALL,
):
    if label is None:
        return
    if axis == "x":
        axes.set_xlabel(label, fontdict={"fontsize": font_size})
    if axis == "y":
        axes.set_ylabel(label, fontdict={"fontsize": font_size})


def create_figure_axes(
    figure: t.Optional[plt.figure] = None,
    axes: t.Optional[plt.axes] = None,
    figure_size: t.Sequence = None,
    font_size: t.Optional[int] = 10,
    aspect: str = "auto",
) -> t.Tuple[plt.figure, plt.axes]:
    """
    Creates figure with axes and sets font size

    Parameters:
    ----------
    fig: Figure if available
    ax: Axes if available
    figure_size: Size of figure (height, width), default (10, 6)
    font_size: Font size

    Returns
    -------
    matplotlib figure and axes
    """
    if font_size is not None:
        set_font(font_size=font_size)
    if figure_size is None:
        figure_size = (10, 6)
    if figure is None and axes is None:
        figure = plt.figure(figsize=figure_size)
        axes = plt.gca()
        axes.set_aspect(aspect)
    if axes is None:
        axes = plt.gca()
        axes.set_aspect(aspect)
    if figure is None:
        figure = axes.get_figure()
    return figure, axes


def create_axes_grid(
    n_columns: int,
    n_rows: int,
    figure_size: t.Optional[t.List[float]] = None,
    widths_along_x: t.Optional[t.List] = None,
    heights_along_y: t.Optional[t.List] = None,
    top: float = 0.05,
    bottom: float = 0.05,
    right: float = 0.05,
    left: float = 0.05,
    spacing_x: float = 0.05,
    spacing_y: float = 0.05,
    spacing_colorbar: float = 0.04,
    colorbar_width: float = 0.07,
    colorbar_skip_row_col: t.Optional[t.List] = None,
    colorbar_include_row_columns: t.Optional[t.List] = None,
    colorbar_off: bool = True,
    skip_columns: t.Optional[t.List] = None,
    skip_rows: t.Optional[t.List] = None,
    skip_row_column: t.Optional[t.List] = None,
    unravel: bool = False,
) -> t.Tuple[plt.figure, np.ndarray, np.ndarray]:
    """
    Create figure with grid of axes. Units are scaled to normalized figure size (max value: 1)
    unless specified otherwise.
    See also https://docs.google.com/presentation/d/1Ec-000rszefjCsv_sgUO62eGyT0-YbYzbk1aLQkU2gM/edit?usp=sharing
    Parameters:
    ----------
    n_cols: Number of columns
    n_rows: Number of rows
    figure_size: Size of figure
    widths_along_x: Normalized width ratios along horizontal direction
    heights_along_y: Normalized height ratios along vertical direction
    top: Offset of axes grid from the top
    bottom: Offset of axes grid from the bottom
    right: Offset of axes grid from the right
    left: Offset of axes grid from the left
    spacing_x: Spacing between axes along horizontal direction
    spacing_y: Spacing between axes along vertical direction
    spacing_colorbar: Spacing between axes and colorbar axes
    colorbar_width: Width of the colorbars
    colorbar_skip_row_col: (row, col) pairs of colorbars to be skipped
    colorbar_include_row_col: (row, col) pairs of colorbars to be plotted
    colorbar_off: No colorbars plotted
    skip_cols: Leave these columns blank
    skip_rows: Leave these rows blank
    skip_row_col: (row, col) pairs of colorbars for which no axes plotted

    Returns
    -------
    Figure, List of axes, List of colorbar axes
    """
    if figure_size is None:
        figure_size = [10.0, 6.0]
    if colorbar_off and colorbar_include_row_columns is None and colorbar_skip_row_col is None:
        colorbar_skip_row_col = [[j, i] for i in range(n_columns) for j in range(n_rows)]
    fig = _create_figure(figure_size)
    a2.utils.utils.all_same_type([n_columns, n_rows], int)
    a2.utils.utils.all_same_type(
        [top, bottom, right, left, spacing_x, spacing_y, spacing_colorbar, colorbar_width], float
    )
    a2.utils.utils.assert_same_type_as(skip_row_column, [[]])
    a2.utils.utils.assert_same_type_as(colorbar_skip_row_col, [[]])
    a2.utils.utils.assert_same_type_as(colorbar_include_row_columns, [[]])
    a2.utils.utils.assert_shape(widths_along_x, (n_columns,), "widths_along_x")
    a2.utils.utils.assert_shape(heights_along_y, (n_rows,), "heights_along_y")

    if skip_rows is None:
        skip_rows = []
    if skip_columns is None:
        skip_columns = []
    if skip_row_column is None:
        skip_row_column = []

    height_total = 1 - (n_rows - 1) * spacing_y - top - bottom
    if heights_along_y is not None:
        heights_along_y = [x / sum(heights_along_y) * height_total for x in heights_along_y]
        axes_heights = np.array([[x for i in range(n_columns)] for x in heights_along_y])
    else:
        height = height_total / n_rows
        axes_heights = np.array([[height for i in range(n_columns)] for j in range(n_rows)])

    colorbar_heights = np.zeros((n_rows, n_columns))
    colorbar_widths = np.zeros((n_rows, n_columns))
    spacings_colorbar = np.zeros((n_rows, n_columns))
    skip_col_colorbar = []
    if colorbar_include_row_columns is not None:
        include_col_colorbar = [x[1] for x in colorbar_include_row_columns]
        skip_col_colorbar = [x for x in range(n_columns) if x not in include_col_colorbar]
    if colorbar_skip_row_col is not None:
        skip_col = [x[1] for x in colorbar_skip_row_col]
        counts = collections.Counter(skip_col)
        for k, v in counts.items():
            if v == n_rows:
                skip_col_colorbar.append(k)
    for i_col in range(n_columns):
        for i_row in range(n_rows):
            if i_col in skip_col_colorbar:
                continue
            colorbar_heights[i_row, i_col] = axes_heights[i_row, i_col]
            colorbar_widths[i_row, i_col] = colorbar_width
            spacings_colorbar[i_row, i_col] = spacing_colorbar
    width_total = (
        1
        - (n_columns - 1) * spacing_x
        - left
        - right
        - max(sum(colorbar_widths[i, :]) for i in range(n_rows))
        - max(sum(spacings_colorbar[i, :]) for i in range(n_rows))
    )
    if widths_along_x is not None:
        widths_along_x = [x / sum(widths_along_x) * width_total for x in widths_along_x]
        axes_widths = np.array([widths_along_x for x in range(n_rows)])
    else:
        width = width_total / n_columns
        axes_widths = np.array([[width for i in range(n_columns)] for j in range(n_rows)])

    axes = [[None for i in range(n_columns)] for j in range(n_rows)]
    axes_colorbar = [[None for i in range(n_columns)] for j in range(n_rows)]
    for i_col in range(n_columns):
        if i_col in skip_columns:
            continue
        for i_row in range(n_rows):
            if i_row in skip_rows:
                continue
            if pair_in_list([i_row, i_col], skip_row_column):
                continue
            axes_left, axes_bottom, axes_width, axes_height = _determine_axes_dimensions(
                n_rows,
                bottom,
                left,
                spacing_x,
                spacing_y,
                axes_heights,
                colorbar_widths,
                spacings_colorbar,
                i_col,
                i_row,
                axes_widths,
            )
            axes[i_row][i_col] = plt.axes([axes_left, axes_bottom, axes_width, axes_height])
            if colorbar_skip_row_col is not None and pair_in_list([i_row, i_col], colorbar_skip_row_col):
                continue
            if colorbar_include_row_columns is not None and not pair_in_list(
                [i_row, i_col], colorbar_include_row_columns
            ):
                continue
            colorbar_width, colorbar_left, colorbar_bottom, colorbar_height = _determine_colorbar_axes_dimensions(
                colorbar_heights, colorbar_widths, spacings_colorbar, i_col, i_row, axes_left, axes_bottom, axes_width
            )
            axes_colorbar[i_row][i_col] = plt.axes([colorbar_left, colorbar_bottom, colorbar_width, colorbar_height])
    axes, axes_colorbar = np.array(axes, dtype=object), np.array(axes_colorbar, dtype=object)
    if unravel:
        axes = a2.utils.utils.flatten_list(axes)
        axes_colorbar = a2.utils.utils.flatten_list(axes_colorbar)
        if n_columns * n_rows == 1:
            axes = axes[0]
            axes_colorbar = axes_colorbar[0]
    return fig, axes, axes_colorbar


def pair_in_list(pair, _list):
    return tuple(pair) in _list or list(pair) in _list


def _determine_colorbar_axes_dimensions(
    colorbar_heights, colorbar_widths, spacings_colorbar, i_col, i_row, axes_left, axes_bottom, axes_width
):
    colorbar_left = axes_left + axes_width + spacings_colorbar[i_row][i_col]
    colorbar_bottom = axes_bottom
    colorbar_width = colorbar_widths[i_row, i_col]
    colorbar_height = colorbar_heights[i_row, i_col]
    return colorbar_width, colorbar_left, colorbar_bottom, colorbar_height


def _determine_axes_dimensions(
    n_rows,
    bottom,
    left,
    spacing_x,
    spacing_y,
    axes_heights,
    colorbar_widths,
    spacings_colorbar,
    i_col,
    i_row,
    axes_widths,
):
    axes_heights_cumulative = axes_heights[slice(i_row + 1, None), i_col]
    axes_left = (
        left
        + sum(axes_widths[i_row, 0:i_col])
        + sum(colorbar_widths[i_row, 0:i_col])
        + sum(spacings_colorbar[i_row, 0:i_col])
        + i_col * spacing_x
    )
    axes_bottom = bottom + sum(axes_heights_cumulative) + (n_rows - i_row - 1) * spacing_y
    axes_width = axes_widths[i_row, i_col]
    axes_height = axes_heights[i_row, i_col]
    return axes_left, axes_bottom, axes_width, axes_height


def annotate_values(
    H: np.ndarray,
    axes: plt.axes,
    size_x: int,
    size_y: int,
    color: str = "black",
    round_to_base: int | None = None,
    font_size: int | None = None,
):
    """
    Overplot values on plot based on 2d matrix whose values are
    plotted in equidistant intervals

    Parameters:
    ----------
    H: Matrix (2d) whose values will be overplot
    ax: Matplotlib axes
    size_x: Number of elements along x-axis
    size_y: Number of elements along y-axis

    Returns
    -------
    matplotlib figure and axes
    """
    x_start = 0
    x_end = 1
    y_start = 0
    y_end = 1
    jump_x = (x_end - x_start) / (2.0 * size_x)
    jump_y = (y_end - y_start) / (2.0 * size_y)
    x_positions = np.linspace(start=x_start, stop=x_end, num=size_x, endpoint=False)
    y_positions = np.linspace(start=y_start, stop=y_end, num=size_y, endpoint=False)
    H_processed = H.copy()
    if round_to_base is not None:
        H_processed = np.round(H, round_to_base)
        if round_to_base < 0:
            H_processed = np.array(H_processed, int)
    for x_index, x in enumerate(x_positions):
        for y_index, y in enumerate(y_positions):
            label = H_processed[y_index, x_index]
            text_x = x + jump_x
            text_y = y + jump_y
            text_y = 1 - text_y
            axes.text(
                text_x,
                text_y,
                label,
                color=color,
                ha="center",
                va="center",
                transform=axes.transAxes,
                fontsize=font_size,
            )


def set_axis_tick_labels(
    axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES,
    values: Sequence[float] | np.ndarray | None = None,
    labels: Sequence | np.ndarray | None = None,
    date_formatter: str | None = None,
    axis: str = "x",
    rotation: int = 0,
    font_size: float = a2.utils.constants.DEFAULT_FONTSIZE_SMALL,
    max_tick_labels: int | None = None,
) -> a2.utils.constants.TYPE_MATPLOTLIB_AXES:
    """
    Set new tick labels for given values

    Parameters:
    ----------
    axes:
        Matplotlib axes
    values:
        Values for corresponding new labels (should also set labels)
    labels:
        New labels for corresponding values (should also set values)
    date_formatter:
        Format string to datetime values as tick labels, e.g. "%Y-%m-%d"
        See https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior for format codes.
    axis:
        Axis to set, i.e. "x" or "y"
    rotation:
        Angle of rotation of tick labels
    font_size:
        Fontsize of axis tick labels
    max_tick_labels:
        Maximum number of tick labels

    Returns
    -------
    axes
    """
    if axis == "x":
        if values is not None:
            axes.set_xticks(values)
        if labels is not None:
            axes.set_xticklabels(labels)
        if date_formatter is not None:
            axes.xaxis.set_major_formatter(matplotlib.dates.DateFormatter(date_formatter))
        if max_tick_labels is not None:
            axes.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_tick_labels))
        axes.tick_params(axis="x", labelrotation=rotation, labelsize=font_size)
    elif axis == "y":
        if values is not None:
            axes.set_yticks(values)
        if labels is not None:
            axes.set_yticklabels(labels)
        if date_formatter is not None:
            axes.yaxis.set_major_formatter(matplotlib.dates.DateFormatter(date_formatter))
        if max_tick_labels is not None:
            axes.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_tick_labels))
        axes.tick_params(axis="y", labelrotation=rotation, labelsize=font_size)
    return axes


def remove_tick_labels(ax: plt.axes, axis: str = "x"):
    """Remove ticks and tick labels for specified axis"""
    set_axis_tick_labels(axes=ax, values=[], labels=[], axis=axis)


def _get_values_from_bar_object(
    bar_object: matplotlib.container.BarContainer,
) -> t.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Debug function to obtain plotted values from a bar plot object

    Returns X, Y and H values of original plot
    Parameters:
    ----------
    bar_object: Bar plot object

    Returns
    -------
    X, Y, H
    """
    X = []
    Y = []
    H = []
    for b in bar_object:
        x, y = b.get_xy()
        X.append(x)
        Y.append(y)
        H.append(b.get_height())
    return np.array(X), np.array(Y), np.array(H)


def _get_values_from_pcolormesh_object(
    pc_object: matplotlib.collections.QuadMesh,
) -> np.ndarray:
    """
    Debug function to obtain plotted values from pcolormesh object

    Returns flattened H values
    Parameters:
    ----------
    pc_object: Pcolormesh plot object

    Returns
    -------
    Flattened matrix values
    """
    return pc_object.get_array().data


def to_list(x, n=2):
    if not (isinstance(x, list) or isinstance(x, tuple)):
        return [x] * n
    if len(x) != n:
        raise ValueError(f"{x} doesn't have expected length {n=}")
    return x


def _plot_colorbar(
    plot, cax: a2.utils.constants.TYPE_MATPLOTLIB_AXES | None = None
) -> a2.utils.constants.TYPE_MATPLOTLIB_COLORBAR:
    colorbar = plt.colorbar(plot, cax=cax)
    return colorbar


def plot_colorbar(
    plot: matplotlib.cm.ScalarMappable,
    cax: a2.utils.constants.TYPE_MATPLOTLIB_AXES = None,
    label: str | None = None,
    font_size: float = a2.utils.constants.DEFAULT_FONTSIZE_SMALL,
) -> a2.utils.constants.TYPE_MATPLOTLIB_COLORBAR:
    colorbar = _plot_colorbar(plot, cax=cax)
    ax_colorbar = colorbar.ax
    if label is not None:
        ax_colorbar.set_ylabel(label, fontdict={"fontsize": font_size})
    set_axis_tick_labels(ax_colorbar, font_size=font_size, axis="y")
    return colorbar


def get_norm(
    norm: str | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    norm_symlog_linear_threshold: float | None = None,
) -> matplotlib.colors.LogNorm | matplotlib.colors.Normalize:
    """
    Returns matplotlib norm used for normalizing matplotlib's colorbars

    Parameters:
    ----------
    norm:
        Normalization type (None/"linear", "log", "symlog")
    vmin:
        Minimum value of normalized colors
    vmax:
        Maximum value of normalized colors
    norm_symlog_linear_threshold:
        Threshold value below which symlog of norm becomes linear.

    Returns
    -------
    Norm object from matplotlib.colors
    """
    if norm == "log":
        return matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax)
    elif norm == "symlog":
        if norm_symlog_linear_threshold is None:
            raise ValueError(f"{norm_symlog_linear_threshold=} needs to specified to use {norm=}")
        return matplotlib.colors.SymLogNorm(vmin=vmin, vmax=vmax, linthresh=norm_symlog_linear_threshold)
    elif norm == "linear" or norm is None:
        return matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
    else:
        raise ValueError(f"Norm: {norm} unknown!")



from typing import Literal

import a2.training.training_hugging
import a2.utils.utils
import transformers


SUPPORTED_MODELS: Literal = ["deberta_base", "deberta_small", "electra_base"]
SUPPORTED_TRAINERS: Literal = ["default", None]
SUPPORTED_LOSSES: Literal = ["default_loss", "focal_loss"]


def update_hyperparameters():

    return


def get_model_config(model_name: SUPPORTED_MODELS, parameters_overwrite: dict = None):
    """
    Get data class that holds hyper parameters of models that are specified via their name (str)

    Parameters:
    ----------
    model_name: Name of model, only selection available, see `a2.training.model_configs.SUPPORTED_MODELS`
    parameters_overwrite: Overwrites hyper parameters based on dictionary key-value pairs,
        ignores keys not present in original data class

    Returns
    -------
    dictionary of metrics
    """
    if parameters_overwrite is None:
        parameters_overwrite = {}
    if model_name == "deberta_base" or model_name == "deberta_small":
        hyper_parameters = a2.training.training_hugging.HyperParametersDebertaClassifier()
    elif model_name == "electra_base":
        hyper_parameters = a2.training.training_hugging.HyperParametersElectraClassifier()
    else:
        raise ValueError(f"{model_name=} not supported, ({SUPPORTED_MODELS=})!")
    hyper_parameters.update(parameters_overwrite)
    return hyper_parameters


def get_customized_trainer_class(trainer_name: SUPPORTED_TRAINERS, method_overwrites: list | None = None):
    if method_overwrites is None:
        method_overwrites = []
    if trainer_name == "default" or trainer_name is None:
        trainer = transformers.Trainer
    else:
        raise ValueError(f"{trainer_name=} not supported ({SUPPORTED_MODELS=})!")

    if a2.utils.utils.is_in_list_and_remove("focal_loss", method_overwrites):
        trainer.compute_loss = a2.training.training_hugging.TrainerWithFocalLoss.compute_loss
    elif a2.utils.utils.is_in_list_and_remove("default_loss", method_overwrites):
        pass
    if len(method_overwrites) != 0:
        raise ValueError(f"{method_overwrites=} not supported ({SUPPORTED_LOSSES=})!")

    return trainer



import collections.abc
import csv
import fnmatch
import glob
import json
import logging
import os
import pathlib
import tarfile
import typing as t
import warnings
import zipfile

import requests
import tqdm


def csv_append(filename: t.Union[str, pathlib.Path], row: list) -> None:
    """
    append row to csv file

    Parameters:
    ----------
    filename: path to file
    row: row to add to file

    Returns
    -------
    """
    with open(filename, "a", newline="", encoding="utf-8") as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(row)


def json_dump(
    filename: t.Union[str, pathlib.Path],
    data: object,
    check_is_file: bool = False,
    log_if_new_file: bool = False,
) -> None:
    """
    dump data to json file

    Parameters:
    ----------
    filename: path to file
    data: data to dump
    check_is_file: if True, only writes file if it doesn't exist
    log_if_new_file: if True, logs creation of new file

    Returns
    -------
    """
    if not check_is_file or (check_is_file and not os.path.isfile(filename)):
        if not os.path.isfile(filename) and log_if_new_file:
            logging.info(f"initializing new file: {filename}")
        with open(filename, mode="w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)


def json_load(filename: t.Union[str, pathlib.Path], raise_if_none: bool = True) -> list:
    """
    load json file

    Parameters:
    ----------
    filename: path to file

    Returns
    -------
    """
    if raise_if_none and not os.path.isfile(filename):
        raise ValueError(f"can't load file: {filename} as doesn't exist!")
    return json.loads(open(filename).read())


def csv_create(
    filename: t.Union[str, pathlib.Path],
    header: t.Iterable[str] | None = None,
    row: t.Iterable[str] | None = None,
    check_is_file_or_empty: bool = True,
) -> None:
    """
    initializes csv file, adds header, row (optional)

    Parameters:
    ----------
    filename: path to file
    header: header as list of strings
    row: row to add to file

    Returns
    -------
    """
    if not check_is_file_or_empty or (
        check_is_file_or_empty and (not os.path.isfile(filename) or not os.path.getsize(filename))
    ):
        logging.info(f"creating new csv file: {filename}")
        with open(filename, "w", newline="", encoding="utf-8") as fp:
            if header is not None:
                for h in header:
                    fp.write(h)
            csv_writer = csv.writer(fp, escapechar="\\")
            if row is not None:
                csv_writer.writerow(row)


def get_header(filepath: t.Union[str, pathlib.Path], n: int = 2) -> list[str]:
    """
    return header of file where header is n lines long

    Parameters:
    ----------
    filepath: path to file
    n: number of rows covering header

    Returns
    -------
    header of file
    """
    header = []
    with open(filepath) as f:
        for i in range(n):
            header.append(f.readline())
    return header


def csv_open(filename: t.Union[str, pathlib.Path]) -> list:
    """
    append row to csv file

    Parameters:
    ----------
    filename: path to file
    row: row to add to file

    Returns
    -------
    """
    with open(filename) as read_obj:
        csv_reader = csv.reader(read_obj)
        return [row[0] if len(row) == 1 else row for row in csv_reader]


def download_file(url: str, folder: str | None = None, overwrite: bool = False) -> str:
    """
    Download file based on url to folder

    File with same name can be overwritten by setting `overwrite` to True
    Parameters:
    ----------
    url: url of file
    folder: folder where file saved
    overwrite: overwrite file with same name by downloaded file

    Returns
    -------
    """
    local_filename = ""
    if folder is not None:
        local_filename = folder
    local_filename = local_filename + url.split("/")[-1]
    if os.path.isfile(local_filename) and not overwrite:
        logging.info(f"file: {local_filename} already exists, no download required.")
        return local_filename
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in tqdm.tqdm(r.iter_content(chunk_size=8192)):
                f.write(chunk)
    return local_filename


def unzip_file(
    path_to_zip_file: t.Union[str, pathlib.Path],
    directory_to_extract_to: t.Union[str, pathlib.Path],
) -> None:
    """Unzip file to directory"""
    with zipfile.ZipFile(path_to_zip_file, "r") as zip_ref:
        zip_ref.extractall(directory_to_extract_to)


def decompress_tarball(filename: t.Union[str, pathlib.Path], to_folder: t.Union[str, pathlib.Path]):
    """Decompress tarball to folder"""
    with tarfile.open(filename) as tar:
        tar.extractall(path=to_folder)


def get_all_files(
    path: t.Union[str, pathlib.Path],
    include_subdirectories: bool = False,
    recursive: bool = False,
):
    """
    Find files based on path pattern.

    Parameters:
    ----------
    path: Base path pattern, which will be used to search directory
    include_subdirectories: Include sub directories in search
    recursive: Make a recursive search (if include_subdirectories=False)

    Returns
    -------
    List of filenames
    """
    path = pathlib.Path(path)
    if include_subdirectories:
        all_files = []
        root = path.parent
        pattern = path.name
        for current_path, subdirs, files in os.walk(root):
            for name in files:
                if fnmatch.fnmatch(name, pattern):
                    all_files.append(os.path.join(current_path, name))
    else:
        path = str(path)
        all_files = glob.glob(path, recursive=recursive)
    return all_files


def remove_existing_files(files_wildcards: t.Union[str, list[str]], include_subdirectories: bool = False) -> None:
    """Remove files based on wildcards. Optionally include subdirectories in search for pattern."""
    if not isinstance(files_wildcards, collections.abc.Iterable) or isinstance(files_wildcards, str):
        files_wildcards = [files_wildcards]
    for fw in files_wildcards:
        files = get_all_files(fw, include_subdirectories=include_subdirectories)
        for f in files:
            if os.path.isfile(f):
                os.remove(f)


def file_exists_and_not_empty(filename: t.Union[str, pathlib.Path]) -> bool:
    """Check if file exists and is not empty"""
    return os.path.exists(filename) and os.path.getsize(filename) != 0


def sort_filenames_by_pattern(file_list: list, pattern_extraction_function: t.Callable) -> list:
    return sorted(
        file_list,
        key=lambda x: pattern_extraction_function(os.path.split(x)[1]),
    )


def is_jsc():
    """Check if currently working on the JSC supercomputer"""
    return os.path.isdir("/p/scratch/deepacf")


def get_folder_data():
    """Get the name of the folder where data for this project is stored indepent of machine used"""
    if is_jsc():
        folder_data = pathlib.Path("/p/project/deepacf/maelstrom/ehlert1/data/")
    else:
        folder_data = pathlib.Path("/home/kristian/Projects/a2/data/")
    if not os.path.isdir(folder_data):
        warnings.warn(f"{folder_data=} does not exist!")
    return folder_data


def get_folder_models():
    """Get the name of the folder where models for this project are stored indepent of machine used"""
    if is_jsc():
        folder_models = pathlib.Path("/p/scratch/deepacf/maelstrom/maelstrom_data/ap2/models_output/")
    else:
        folder_models = pathlib.Path("/home/kristian/Projects/a2/models/")
    if not os.path.isdir(folder_models):
        warnings.warn(f"{folder_models=} does not exist!")
    return folder_models


def get_folder_radar(folder_radar: t.Optional[pathlib.Path] = None) -> pathlib.Path:
    """Get folder where radar data is saved"""
    if folder_radar is None:
        folder_radar = get_folder_data() / "precipitation/radar/"
    return folder_radar


def check_acess_rights_folder(file_or_folder):
    if os.path.isdir(file_or_folder):
        description = "files in folder"
    else:
        description = "file"
    if os.access(file_or_folder, os.X_OK):
        print(f"User can execute {description} {file_or_folder}")
    else:
        print(f"User cannot execute {description} {file_or_folder}")
    if os.access(file_or_folder, os.W_OK):
        print(f"User can write {description} {file_or_folder}")
    else:
        print(f"User cannot write {description} {file_or_folder}")
    if os.access(file_or_folder, os.R_OK):
        print(f"User can read {description} {file_or_folder}")
    else:
        print(f"User cannot read {description} {file_or_folder}")


def folder_exists(path: str | pathlib.Path, check_if_empty: bool = False, raise_exception: bool = False) -> bool:
    if not os.path.isdir(path):
        if raise_exception:
            raise ValueError(f"{path=} doesn't exist!")
        return False
    if check_if_empty and os.path.getsize(path) <= 4096:
        if raise_exception:
            raise ValueError(f"{path=} is empty!")
        return False
    return True


def make_directories(path: str | pathlib.Path):
    if not os.path.isdir(path):
        logging.info(f"... making {path=}")
        os.makedirs(path)



import a2.training.evaluate_hugging
import a2.training.model_configs
import a2.training.tracking
import a2.training.tracking_hugging
import a2.training.training_deep500
import a2.training.training_hugging



import a2.plotting
import a2.dataset
import numpy as np
import logging
import pathlib
import a2.cli.cli_plotting.plot as plot


FOLDER_DATA = a2.utils.file_handling.get_folder_data()
FOLDER_TWEETS = FOLDER_DATA / "tweets/"
FOLDER_WEATHER_STATIONS = FOLDER_DATA / "weather_stations/"
FOLDER_FIGURES = pathlib.Path('../../figures/')


@plot.cli.command("map")
def plot_map(
    filename_tweets=FOLDER_TWEETS
    / "tweets_no_keywords_2020-02-13T00:00:00.000Z_2020-02-14T00:00:00_locations_bba_era5.nc",
    filename_weather_stations=FOLDER_WEATHER_STATIONS / "weather_stations_hourly_rainfall_uk_2017-2020_station_number.csv",
    starting_time="2020-02-13T03:30:00.000000000",
    filename="test.pdf",
):
    import matplotlib
    matplotlib.use('PDF')
    
    logging.info(f'... loading {filename_tweets}')
    ds_tweets = a2.dataset.load_dataset.load_tweets_dataset(filename_tweets)
    ds_tweets['tp_era5_mm'] = (['index'], ds_tweets.tp_era5.values * 1e3)
    logging.info(f'... loading {filename_weather_stations}')
    df_stations = a2.dataset.load_dataset.load_weather_stations(filename_weather_stations)
    logging.info(f'... plotting')
    key_tweets = a2.dataset.tweets.KeyTweets(station_tp="tp_era5_mm")
    figure = a2.plotting.weather_maps.plot_tp_station_tweets(
        ds_tweets,
        df_stations,
        grid_shape=(5, 4),
        colormap="plasma",
        vmin=0,
        vmax=1,
        fontsize=14,
        choice_type="increment_time",
        selection_delta_time=1,
        selection_delta_time_units="h",
        selector_use_limits=[True, False],
        increment_time_delta=1,
        increment_time_delta_units="h",
        increment_time_value=np.datetime64(starting_time),
        # xlim=[-5, 0],
        # ylim=[51, 56],
        xlim=[-0.5, 0.5],
        ylim=[51, 52],
        processes=10,
        key_tweets=key_tweets,
        circle_size=0.03,
    )
    filename = FOLDER_FIGURES / "maps/stations/" / filename
    logging.info(f"... saving {filename}")
    a2.plotting.figures.save_figure(figure, filename)




import argparse
import pathlib

import a2.training
import a2.utils


def get_parser(formatter_class=argparse.ArgumentDefaultsHelpFormatter):
    parser = argparse.ArgumentParser(formatter_class=formatter_class)
    return parser


def dataset_tweets(parser):
    parser.add_argument(
        "--tweets_dir",
        "-inpath",
        type=str,
        default=a2.utils.file_handling.get_folder_data() / "tweets/",
        help="Directory where input netCDF-files are stored.",
    )


def dataset(parser):
    parser.add_argument(
        "--dataset_prefix",
        type=str,
        default="2020_tweets_rain_sun_vocab_emojis_locations_bba_Tp_era5_no_bots_normalized_filtered_weather_stations_fix_predicted_simpledeberta_radar",  # noqa: E501
        help="Dataset prefix determined during split of dataset.",
    )
    parser.add_argument(
        "--data_description",
        "-datad",
        type=str,
        default="tweets 2017-2020, keywords emojis as description, keywords only",
        help="Data description purely used for logging.",
    )


def evaluation(parser):
    parser.add_argument(
        "--evaluation_strategy", "-estrat", type=str, default="epoch", help="When to evaluate steps/epoch/no"
    )
    parser.add_argument(
        "--eval_steps",
        "-esteps",
        type=int,
        default=None,
        help="Number of steps between evaluations if evaluation strategy is set to steps.",
    )


def classifier(parser):
    parser.add_argument("--key_text", type=str, default="text_normalized", help="Key for text in input file.")
    parser.add_argument(
        "--classifier_domain", choices=["rain", "relevance"], type=str, required=True, help="Domain of classifier"
    )
    parser.add_argument(
        "--key_input",
        type=str,
        default="text",
        choices=["text"],
        help="Column name of dataset that corresponds to input of classifier.",
    )
    parser.add_argument(
        "--key_output",
        type=str,
        default="raining",
        choices=["raining", "relevant", "raining_station"],
        help="Column name of dataset that corresponds to output of classifier.",
    )


def dataset_training(parser):
    parser.add_argument(
        "--filename_dataset_train", required=True, type=str, help="Filename of dataset used for training."
    )
    parser.add_argument(
        "--filename_dataset_validate", required=True, type=str, help="Filename of dataset used for validation."
    )
    parser.add_argument(
        "--filename_dataset_test", required=True, type=str, help="Filename of dataset used for testing."
    )


def output(parser):
    parser.add_argument(
        "--output_dir",
        type=str,
        default="output/",
        help="Folder to save results.",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        required=True,
        help="Name of task, results will be saved in this subdirectory of output.",
    )


def dataset_rain(parser):
    parser.add_argument("--key_rain", type=str, default="tp_h", help="Key for rain data in input file.")
    parser.add_argument(
        "--threshold_rain",
        type=float,
        default=1e-8,
        help="Threshold above which precipitation is considered raining in input file.",
    )


def dataset_weather_stations(parser):
    parser.add_argument(
        "--kms_within_station",
        "-kms",
        type=float,
        default=1,
        help="Distance to weather station. Used to create weather station dataset.",
    )
    parser.add_argument(
        "--key_distance_weather_station",
        type=str,
        default="station_distance_km",
        help="Key in dataset that shows distance to weather station. Used to create weather station dataset.",
    )
    parser.add_argument(
        "--weather_station_dataset_prefix",
        type=str,
        default="WeatherStationDataset",
        help="Prefix of weather station dataset.",
    )


def dataset_relevance(parser):
    parser.add_argument(
        "--key_relevance",
        type=str,
        default="relevant",
        help="Variable name that specifies relevance of Tweet.",
    )


def dataset_relevance_split(parser):
    parser.add_argument(
        "--filename_tweets_with_keywords",
        type=str,
        default="2017_2020_tweets_rain_sun_vocab_emojis_locations_bba_Tp_era5_no_bots_normalized_filtered.nc",
        help="Filename of training data.",
    )
    parser.add_argument(
        "--data_filename_irrelevant",
        type=str,
        default="2017_2020_tweets_rain_sun_vocab_emojis_locations_bba_Tp_era5_no_bots_normalized_filtered.nc",
        help="Filename of training data.",
    )
    parser.add_argument(
        "--n_tweets_irrelevant",
        type=int,
        default=-1,
        help="Number of irrelevant tweets used to create dataset, "
        "use all irrelevant tweets by default (`n_tweets_irrelevant=-1`).",
    )
    parser.add_argument(
        "--split_dir",
        type=str,
        default="split_relevance",
        help="Folder to save split datasets.",
    )
    parser.add_argument(
        "--raining_classifier_dataset_prefix",
        type=str,
        default="RainingClassifierDataset",
        help="Prefix of raining classifier dataset.",
    )
    parser.add_argument(
        "--relevance_classifier_dataset_prefix",
        type=str,
        default="RelevanceClassifierDataset",
        help="Prefix of relevance classifier dataset.",
    )


def dataset_split_sizes(parser):
    parser.add_argument("--test_size", "-ts", type=float, default=0.2, help="Fraction of test set.")
    parser.add_argument("--validation_size", "-vs", type=float, default=0.2, help="Fraction of validation set.")


def dataset_split(parser):
    dataset_split_sizes(parser)
    parser.add_argument(
        "--filename_dataset_to_split",
        required=True,
        type=pathlib.Path,
        help="Filename of dataset that will be split.",
    )
    parser.add_argument(
        "--key_stratify",
        type=str,
        default=None,
        help="Key used to stratify dataset split.",
    )


def dataset_select(parser):
    parser.add_argument(
        "--select_relevant",
        type=custom_boolean,
        default=True,
        help="Whether to only select Tweets as revelant (`args.key_relevance`==True).",
    )


def predict_model(parser):
    parser.add_argument(
        "--filename_dataset_predict",
        required=True,
        type=pathlib.Path,
        help="Filename of dataset that predictions will be made for.",
    )
    parser.add_argument(
        "--path_raw_model",
        type=str,
        required=True,
        help="Directory where untrained model saved.",
    )
    parser.add_argument(
        "--path_trained_model",
        type=str,
        required=True,
        help="Directory where trained model saved.",
    )


def model(parser):
    parser.add_argument(
        "--model_path",
        "-in",
        type=str,
        default="/p/project/deepacf/maelstrom/ehlert1/deberta-v3-small/",
        help="Directory where input netCDF-files are stored.",
    )
    parser.add_argument(
        "--model_name",
        type=str,
        choices=a2.training.model_configs.SUPPORTED_MODELS,
        default="deberta_small",
        help="Name of model, sets default hyper parameters.",
    )
    parser.add_argument(
        "--output_dir_model",
        "-outdir",
        type=str,
        default=a2.utils.file_handling.get_folder_models(),
        help="Output directory where model is saved.",
    )
    parser.add_argument(
        "--trainer_name",
        type=str,
        default="default",
        choices=a2.training.model_configs.SUPPORTED_TRAINERS,
        help="Trainer class selected by its name.",
    )
    parser.add_argument(
        "--loss",
        type=str,
        default="default_loss",
        choices=a2.training.model_configs.SUPPORTED_LOSSES,
        help="Loss used for training selected by its name.",
    )


def model_classifier(parser):
    parser.add_argument(
        "--num_labels",
        type=int,
        default=2,
        help="Number of labels in classification task.",
    )


def run(parser):
    parser.add_argument(
        "--run_folder", type=str, required=True, help="Output folder where model is saved in `output_dir`."
    )


def mlflow(parser):
    figures(parser)
    parser.add_argument("--run_name", type=str, default="RUNNAME", help="Name of run used for logging only.")
    parser.add_argument(
        "--mlflow_experiment_name",
        type=str,
        default="maelstrom-a2-train",
        help="Name MLflow experiment where results are logged.",
    )


def figures(parser):
    parser.add_argument(
        "--figure_folder", type=str, default="figures/", help="Directory where input netCDF-files are stored."
    )


def hyperparameter_basic(parser):
    parser.add_argument("--random_seed", "-rs", type=int, default=42, help="Random seed value.")
    parser.add_argument("--job_id", "-jid", type=int, default=None, help="Job id when running on hpc.")
    parser.add_argument("--iteration", "-i", type=int, default=0, help="Iteration number when running benchmarks.")


def hyperparameter(parser):
    hyperparameter_basic(parser)
    # arg name should match name of hyper parameter used in `model_configs.py`!
    parser.add_argument("--epochs", type=int, default=1, help="Numer of epochs to train.")
    parser.add_argument("--batch_size", type=int, default=32, help="Number of samples per mini-batch.")
    parser.add_argument("--learning_rate", type=float, default=3e-05, help="Learning rate to train model.")
    parser.add_argument("--weight_decay", type=float, default=0, help="Weight decay rate to train model.")
    parser.add_argument("--warmup_ratio", type=float, default=0, help="Warmup ratio to train model.")
    parser.add_argument("--warmup_steps", type=float, default=500, help="Warmup steps to train model.")
    parser.add_argument(
        "--hidden_dropout_prob", "-hdp", type=float, default=0.1, help="Probability of hidden droup out layer."
    )
    parser.add_argument("--cls_dropout", "-cd", type=float, default=0.1, help="Dropout probability in classifier head.")
    parser.add_argument("--lr_scheduler_type", "-lst", type=str, default="linear", help="Learning rate scheduler type.")

    parser.add_argument(
        "--base_model_weights_fixed",
        "-weightsfixed",
        action="store_true",
        help="Weights of base model (without classification head) are fixed (not trainable).",
    )
    parser.add_argument("--save_steps", type=int, default=500, help="Steps after which model is saved.")
    parser.add_argument("--logging_steps", type=int, default=1, help="Steps after which model logs are written.")


def benchmarks(parser):
    parser.add_argument(
        "--log_gpu_memory",
        action="store_true",
        help="Monitor Cuda memory usage.",
    )
    parser.add_argument(
        "--ignore_tracking",
        action="store_true",
        help="Use mantik tracking.",
    )


def debug(parser):
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Whether to toggle debug mode.",
    )


def custom_boolean(boolean: str) -> bool:
    """Custom argparse type for booleans"""
    if boolean == "False" or boolean == "false" or boolean == "0":
        return False
    elif boolean == "True" or boolean == "true" or boolean == "1":
        return True
    else:
        raise argparse.ArgumentTypeError(f"Couldn't convert {boolean=} to boolean!")



import logging
import operator
import pathlib
import typing as t

import numpy as np


FILE_LOC = pathlib.Path(__file__).parent
EMBEDDINGS_FOLDER = FILE_LOC / "../../../data/embeddings/"
TWITTER_EMBEDDINGS_PATH = EMBEDDINGS_FOLDER / "glove.twitter.27B/"


def load_glove_embeddings(
    twitter_style: bool = True,
    twitter_vector_dimensions: int = 25,
) -> t.Dict:
    """
    Load Glove embeddings

    See https://nlp.stanford.edu/projects/glove/ for more info.
    Parameters:
    ----------
    twitter_style: Use embeddings trained on Twitter based corpus
                   else load love.840B.300d trained on Common Crawl
    twitter_vector_dimensions: Length of vectors representation
                               of tokens 25, 50, 100, 200 available

    Returns
    -------
    glove embeddings as dictionary
    """
    if twitter_style:
        path = TWITTER_EMBEDDINGS_PATH / f"glove.twitter.27B.{twitter_vector_dimensions}d.txt"
        glove_embeddings = {}
        with open(
            path,
            encoding="utf8",
        ) as f:
            for line in f:
                values = line.split()
                word = values[0]
                embedding = np.asarray(values[1:], dtype="float64")
                glove_embeddings[word] = embedding
    else:
        path = EMBEDDINGS_FOLDER / "glove.840B.300d.pkl/glove.840B.300d.pkl"
        glove_embeddings = np.load(path, allow_pickle=True)
    logging.info("Loaded %s word vectors." % len(glove_embeddings))
    return glove_embeddings


def build_vocabulary(texts: t.Sequence) -> t.Dict:
    """
    Build vocabulary based on words in sequence of give text

    'Words' are obtained by splitting texts at whitespaces
    Parameters:
    ----------
    texts: Sequence of texts
    Returns
    -------
    dictionary; keys: words in vocab, values: occurence of words in texts
    """

    def generate_tokens(text):
        return text.split()

    texts_tokenized = np.array(list(map(generate_tokens, texts)), dtype=object)
    vocab: t.Dict[str, int] = {}
    for tt in texts_tokenized:
        for token in tt:
            try:
                vocab[token] += 1
            except KeyError:
                vocab[token] = 1
    return vocab


def check_embeddings_coverage(texts, embeddings):
    """
    Check coverage of words in texts in given set of embeddings

    Returns:
        sorted_oov: Sorted dictionary with occurence of
                    words that are not part of the embedding's vocobulary
        vocab_coverage: Percentage of individual words covered
                        by embedding's vocobulary
        text_coverage: Percentage of total text covered by
                       embedding's vocobulary
        vocab: Dictionary of vocab of texts including occurence of terms
    Parameters:
    ----------
    texts: Sequence of texts
    embeddings: Dictionary of embeddings,
                e.g. `a2.preprocess.embedding.load_glove_embeddings`
    Returns
    -------
    sorted_oov, vocab_coverage, text_coverage, vocab
    """
    vocab = build_vocabulary(texts)
    covered = {}
    oov = {}
    n_covered = 0
    n_oov = 0
    for word in vocab:
        if word == "":
            logging.info("found it")
        try:
            covered[word] = embeddings[word]
            n_covered += vocab[word]
        except KeyError:
            oov[word] = vocab[word]
            n_oov += vocab[word]
    vocab_coverage = len(covered) / len(vocab)
    text_coverage = n_covered / (n_covered + n_oov)
    sorted_oov = sorted(oov.items(), key=operator.itemgetter(1))[::-1]
    return sorted_oov, vocab_coverage, text_coverage, vocab



import re


def is_emoji(string):
    emoji_pattern = re.compile(
        "["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U00002500-\U00002BEF"  # chinese char + emojis?
        "]+",
        re.UNICODE,
    )
    return bool(re.findall(emoji_pattern, string))


def get_emoji_filename(emoji):
    if len(emoji) == 1:
        return f"{ord(emoji):x}.png"
    elif len(emoji) > 1:
        out = ""
        for c in emoji:
            out += f"{ord(c):x}_"
        return f"{out[:-1]}.png"



import numpy as np


def validate_array(x: np.ndarray, type: str = "float", name: str = None):
    """Validate numpy based on `type`, e.g. no nan-values"""
    if type == "float":
        if np.sum(np.isnan(x.astype(float))) > 0:
            raise ValueError(f"Found nan-values {f'in {name}' if name is not None else ''}!")


def assert_type(variable, type_to_check):
    if not isinstance(variable, type_to_check):
        raise TypeError(f"{variable} is of type: {type(variable)} != expected type: {type_to_check}")
